#! /usr/bin/env python3

from sys import argv
from subprocess import check_output, TimeoutExpired

from tkinter import *
from tkinter.filedialog import *

#
# GLOBAL DECLARATIONS
#

TIMEOUT = 10

window = Tk()
window.title("Connect4")

canvas = Canvas(window, width=800, height=400, background='black', cursor="tcross")

menu = PhotoImage(file="./Ressources/4P_main_menu.png")
game = PhotoImage(file="./Ressources/4P_game.png")

p1w = PhotoImage(file="./Ressources/P1_wins.png")
p2w = PhotoImage(file="./Ressources/P2_wins.png")
exaequo = PhotoImage(file="./Ressources/Ex_aequo.png")

instr1 = PhotoImage(file="./Ressources/Instruct1.png")
instr2 = PhotoImage(file="./Ressources/Instruct2.png")

right_arrow = PhotoImage(file="./Ressources/Right_arrow.png")
left_arrow = PhotoImage(file="./Ressources/Left_arrow.png")
down_arrow = PhotoImage(file="./Ressources/Down_arrow.png")

colors = ["ROUGE", "JAUNE"]

i = 0
RA = 0
LA = 0
DA = 0
replay = []
AIs = []
names = []

board = []
board_disp = []

winner = -1
rep_index = 0

why = 0

active_player = 0

players = 2

#
# GUI CODE
#

def previous_turn():
	el = board_disp.pop()
	canvas.delete(el)
	board[replay[rep_index]].pop()

def next_turn(col, player):
	board[col].append(player)
	pion = canvas.create_oval(73 + 40*col, 330 - 40*len(board[col]), 103 + 40 * col, 360 - 40*len(board[col]), fill = "red" if player == 0 else "yellow")
	board_disp.append(pion)

def replay_inputs(event):
	global rep_index

	touche = event.keysym

	if touche == "Escape":
		reboot_menu()
	elif touche == "Left" and rep_index > 0:
		previous_turn()
		rep_index -= 1
	elif touche == "Right" and rep_index < len(replay) - 1:
		rep_index += 1
		next_turn(replay[rep_index], rep_index % 2)
		if rep_index == len(replay) - 1:
			canvas.create_image(10, 10, image = p1w if winner == 0 else (p2w if winner == 1 else exaequo), anchor = NW)
			if why == -1:
				canvas.create_text(570, 10, anchor=NW, text="(timeout)", font="Emulogic 12", fill="grey")
			elif why == -2:
				canvas.create_text(570, 10, anchor=NW, text="(wrong action)", font="Emulogic 12", fill="grey")



def try_to_play():
	global active_player

	if len(board[i]) > 5:
		return
	else:
		next_turn(i, active_player)
		moves = possible_moves()
		if won(active_player):
			winner = active_player
			active_player = 2
			canvas.create_image(10, 10, image = p1w if winner == 0 else p2w, anchor = NW)
			why = 0
			return
		elif moves == []:
			winner = -1
			active_player = 2
			why = 0
			canvas.create_image(10, 10, image = exaequo, anchor = NW)
			return
		else:
			active_player = (active_player + 1) % 2
			if active_player >= players:    
				m = ask_to_play(active_player - players)

				#print(moves, m)

				if m == 'TIMEOUT':
					winner = (active_player+1)%2
					active_player = 2
					why = -1
					return
				else :
					if not (m in moves):
						why = -2
						winner = (active_player + 1) % 2
						active_player = 2
						return
					else:
						next_turn(m, active_player)

						moves = possible_moves()
						if won(active_player):
							winner = active_player
							active_player = 2
							canvas.create_image(10, 10, image = p1w if winner == 0 else p2w, anchor = NW)
							why = 0
							return
						elif moves == []:
							winner = -1
							active_player = 2
							why = 0
							canvas.create_image(10, 10, image = exaequo, anchor = NW)
							return

						active_player = (active_player + 1) % 2


def game_inputs(event):
	global i

	touche = event.keysym

	if touche == "Escape":
		reboot_menu()
	elif touche == "Left" and i > 0:
		i -= 1
	elif touche == "Right" and i < 6:
		i += 1
	elif touche == "Return" and active_player < players:
		try_to_play()

	canvas.coords(DA, 90 + 40*i, 60)


def menu_inputs(event):
	global i
	global players

	touche = event.keysym
	if touche == "Up" and i > 0:
		i -= 1
	elif touche == "Down" and i < 3:
		i += 1

	canvas.coords(RA, 180 if i < 2 else (245 if i < 3 else 220), 70 * (i+1) + (20 if i > 1 else 0))
	canvas.coords(LA, 620 if i < 2 else (555 if i < 3 else 580), 70 * (i+1) + (20 if i > 1 else 0))

	if touche == "Return":
		p_of_i = [1, 2, 0, -1]
		players = p_of_i[i]

		reboot_game()


def reboot_menu():
	global i
	global RA
	global LA

	i = 0

	canvas.create_image(399, 199, image=menu)

	RA = canvas.create_image(180, 70, image=right_arrow)
	LA = canvas.create_image(620, 70, image=left_arrow)

	canvas.focus_set()
	canvas.unbind("<Key>")
	canvas.bind("<Key>", menu_inputs)

	canvas.pack()
	window.mainloop()



def column_repr(col):
    return ' '.join([str(col)] + [ colors[c][0] for c in board[col]])

def board_repr():
    return '\n'.join(column_repr(c) for c in range(7))

def ask_to_play(player):
    s = colors[player] + '\n' + board_repr() + '\n'
    try:
        r = check_output(["python3", AIs[player]],
                         universal_newlines=True,
                         input=s, timeout=TIMEOUT)
        return int(r)
    except TimeoutExpired:
        return 'TIMEOUT'

def align(start, dir, c):
    lin, col = start
    dl, dc = dir
    for k in range(4):
        i = lin + k*dl
        j = col + k*dc
        if not (j in range(7) and i in range(len(board[j])) and board[j][i] == c):
            return False
    return True

def won(c):
    for i in range(6):
        for j in range(7):
            for d in [(0, 1), (1, 0), (1, 1), (1, -1)]:
                if align((i, j), d, c):
                    return True
    return False

def possible_moves():
	return [ i for i in range(7) if len(board[i]) < 6 ]


def calc_replay():
	global board
	global winner
	global why
	global replay

	board = [[] for k in range(7)]
	replay = []

	player = 0
	while True:
		for p in range(2):
			if won(p):
				winner = p
				why = 0
				return

		moves = possible_moves()
		if moves == []:
			winner = -1
			why = 0
			return
            
		m = ask_to_play(player)

		#print(moves, m)

		if m == 'TIMEOUT':
			why = -1
			winner = (player + 1) % 2
			return
		else :
			if not (m in moves):
				why = -2
				winner = (player + 1) % 2
				return
			else:
				board[m].append(player)
				replay.append(m)

				player = (player + 1) % 2


def reboot_game():
	global AIs
	global names
	global replay
	global board
	global winner
	global rep_index
	global why
	global i
	global DA
	global active_player

	AIs = []
	names = []
	winner = -1
	rep_index = 0
	i = 0
	active_player = 0
	why = 0

	if players == -1:
		filename=askopenfilename(title="Select replay",filetypes=[('Text files','.txt'),('All files','.*')])
		with open(filename, 'r') as file:
			read = file.readlines()
			if len(read) < 4:
				return
			names = ["AI : " + read[k].strip().split(".")[0] for k in range(2)]
			replay = [int(col) for col in read[2].strip().split(';')]
			winner = int(read[3].strip().split(';')[0])
			why = int(read[3].strip().split(';')[1])
			print(names, replay, winner, why)
	else:
		for k in range(players):
			names.append("HUMAN")
		for k in range(2-players):
			filename = askopenfilename(title="Select AI "+str(k+1),filetypes=[('Python files','.py'),('All files','.*')])
			AIs.append(filename)
			names.append("AI : " + filename.split("/")[-1].split("\\")[-1].split(".")[0])

		if players == 0:
			calc_replay()
			print(names, replay, winner, why)

	board = [[] for k in range(7)]
	board_disp = []
		
	canvas.unbind("<Key>")

	if players > 0:
		canvas.bind("<Key>", game_inputs)
	else:
		canvas.bind("<Key>", replay_inputs)

	canvas.create_image(399, 199, image=game)
	canvas.create_text(500, 130, anchor=NW, text=names[0], font="Emulogic 12", fill="grey")
	canvas.create_text(500, 280, anchor=NW, text=names[1], font="Emulogic 12", fill="grey")

	if players < 1:
		canvas.create_image(15, 375, anchor = NW, image = instr1)

		if len(replay) > 0:
			next_turn(replay[rep_index], rep_index % 2)
			if rep_index == len(replay) - 1:
				canvas.create_image(10, 10, image = p1w if winner == 0 else (p2w if winner == 1 else exaequo), anchor = NW)
				if why == -1:
					canvas.create_text(570, 10, anchor=NW, text="(timeout)", font="Emulogic 12", fill="grey")
				elif why == -2:
					canvas.create_text(570, 10, anchor=NW, text="(wrong action)", font="Emulogic 12", fill="grey")
		else:
			canvas.create_image(10, 10, image = p2w, anchor = NW)
			if why == -1:
				canvas.create_text(570, 10, anchor=NW, text="(timeout)", font="Emulogic 12", fill="grey")
			elif why == -2:
				canvas.create_text(570, 10, anchor=NW, text="(wrong action)", font="Emulogic 12", fill="grey")

	else:
		canvas.create_image(10, 375, anchor = NW, image = instr2)

		DA = canvas.create_image(90, 60, image = down_arrow)


reboot_menu()




# canvas.create_text(400,70,text = "1-PLAYER START", font = "Emulogic 18", fill="#aaaaff")
# canvas.create_text(400,140,text = "2-PLAYER START", font = "Emulogic 18", fill="#aaaaff")
# canvas.create_text(400,230,text = "AI BATTLE", font = "Emulogic 18", fill="#ffaaaa")
# canvas.create_text(400,300,text = "LOAD REPLAY", font = "Emulogic 18", fill="#ffaaaa")

# canvas.create_text(400,380,text = "Louis Sugy                           CC BY-NC-SA", font = "Emulogic 12", fill="#424242")




# canvas.create_text(580, 100, text = "PLAYER ONE", font = "Emulogic 18", fill="red")
# canvas.create_text(580, 250, text = "PLAYER TWO", font = "Emulogic 18", fill="yellow")
# for k in range(8):
# 	canvas.create_line(70 + 40*k, 80, 70 + 40*k, 330, fill="grey", width=("3" if k==0 or k==7 else "2"))
# canvas.create_line(69, 330, 72 + 40*7, 330, fill="grey", width="3")




# canvas.create_text(0, 0, anchor=NW, text = "EX AEQUO !", font = "Emulogic 16", fill="white")

# canvas.create_text(0, 0, anchor=NW, text = " ENTER : SELECT COLUMN             LEFT/RIGHT : CHANGE COLUMN", font = "Emulogic 10", fill="grey")
# canvas.create_text(0, 0, anchor=NW, text = "LEFT/RIGHT : PREVIOUS/NEXT TURN", font = "Emulogic 10", fill="grey")

# canvas.create_text(0, 0, anchor=NW, text = "V", font = "Emulogic 18", fill="white")


# canvas.pack()
# window.mainloop()
